/* $Id: adr.c,v 1.20 2006/08/07 15:55:51 geni Exp $
 *
 * Copyright (c) 2005 Huidae Cho
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "onhand.h"
#include "define.h"

static void adr_rec_qsort(adr_rec_t *head, adr_rec_t *tail);

int
adr_read(adr_t *db, char *file)
{
	int size, i, j, image;
	u8 *data, *ptr;
	u32 offset;
	adr_rec_t *rec;

	if(read_file(file, &data, &size)){
		ERROR("read_file");
		return -1;
	}
	/* the minimum size is 224 = 4+2+2+27*2*4 */
	if(size < 224 || memcmp(data, pim_magic, 4)){
		ERROR("%s: not an adr file", file);
		return -1;
	}

	adr_init(db);

	for(i = 0; i < 27; i++){
		memcpy(&offset, data+8+i*8, 4);
		while(offset != 0xffffffff && offset < size){
			db->num_recs++;
			ptr = data + offset;
			memcpy(&offset, ptr+4, 4);
			rec = (adr_rec_t *)zalloc(sizeof(adr_rec_t));
			rec->priv = ptr[14];
			if(rec->priv)
				db->num_priv++;
			image = ptr[16];
			ptr += 20;
			for(j = 0; j < ADR_NUM_FIELDS; j++){
				memcpy(&rec->data[j].size, ptr, 2);
				ptr += 2;
				rec->data[j].buf = NULL;
				if(!rec->data[j].size)
					continue;
				rec->data[j].buf =
					(u8 *)malloc(rec->data[j].size+1);
				memcpy(rec->data[j].buf, ptr,
						rec->data[j].size);
				rec->data[j].buf[rec->data[j].size] = 0;
				ptr += rec->data[j].size;
			}
			rec->mmp = NULL;
			if(image){
				u16 width, height;

				memcpy(&width, ptr, 2);
				ptr += 2;
				memcpy(&height, ptr, 2);
				ptr += 2;
				rec->mmp = (mmp_t *)malloc(sizeof(mmp_t));
				mmp_create(rec->mmp, width, height);
				memcpy(rec->mmp->buf, ptr, rec->mmp->width *
						rec->mmp->height / 8);
			}
			rec->prev = rec->next = NULL;
			if(db->head){
				rec->prev = db->tail;
				db->tail->next = rec;
				db->tail = rec;
			}else
				db->head = db->tail = rec;
		}
	}

	free(data);

	adr_rec_sort(db);

	return 0;
}

int
adr_write(adr_t *db, char *file)
{
	int fd, i, j;
	u8 i8;
	u16 i16, size;
	u32 i32, offset, offset2;
	adr_rec_t *rec;

	for(rec = db->head; rec && rec->data[ADR_NAME].buf; rec = rec->next);
	if(rec && !rec->data[ADR_NAME].buf){
		ERROR("internal error");
		return -1;
	}

	_open();

	_write(pim_magic, 4);
	_write(&db->num_priv, 2);
	_write(&db->num_recs, 2);

	/* header */
	offset = 224;
	rec = db->head;
	for(i = 0; i < 27; i++){
		if(!rec || (i == 26 ? 0 :
				tolower(rec->data[ADR_NAME].buf[0]) != 'a'+i)){
			offset2 = 0xffffffff;
			for(j = 0; j < 2; j++)
				_write(&offset2, 4);
			continue;
		}
		_write(&offset, 4);
		for(; rec && (i == 26 ? 1 :
				tolower(rec->data[ADR_NAME].buf[0]) == 'a'+i);
				rec = rec->next){
			offset2 = offset;
			offset += 20;
			for(j = 0; j < ADR_NUM_FIELDS; j++)
				offset += 2 + rec->data[j].size;
			if(rec->mmp)
				offset += 4+rec->mmp->width*rec->mmp->height/8;
		}
		_write(&offset2, 4);
	}

	/* records */
	offset = 224;
	offset2 = 0xffffffff;
	for(rec = db->head; rec; rec = rec->next){
		_write(&offset2, 4);
		offset2 = offset;
		offset += 20;
		size = 0;
		for(j = 0; j < ADR_NUM_FIELDS; j++)
			size += 2 + rec->data[j].size;
		offset += size;
		if(rec->mmp)
			offset += 4 + rec->mmp->width * rec->mmp->height / 8;
		if(rec->next && ((tolower(rec->data[ADR_NAME].buf[0]) ==
				  tolower(rec->next->data[ADR_NAME].buf[0])) ||
				 (!isalpha(rec->data[ADR_NAME].buf[0]) &&
				  !isalpha(rec->next->data[ADR_NAME].buf[0]))))
			_write(&offset, 4);
		else{
			offset2 = 0xffffffff;
			_write(&offset2, 4);
		}
		i32 = 0;
		_write(&i32, 4);

		/* TODO: 00 ... 00 for unknown does not work! */
		i16 = 0xcb5e;
		_write(&i16, 2);

		_write(&rec->priv, 1);
		i8 = 0;
		_write(&i8, 1);
		i8 = (rec->mmp ? 1 : 0);
		_write(&i8, 1);
		i8 = 0;
		_write(&i8, 1);
		_write(&size, 2);
		for(j = 0; j < ADR_NUM_FIELDS; j++){
			_write(&rec->data[j].size, 2);
			if(!rec->data[j].size)
				continue;
			_write(rec->data[j].buf, rec->data[j].size);
		}
		if(rec->mmp){
			_write(&rec->mmp->width, 2);
			_write(&rec->mmp->height, 2);
			j = rec->mmp->width * rec->mmp->height / 8;
			_write(rec->mmp->buf, j);
		}
	}

	_close();

	return 0;
}

void
adr_scan(adr_t *db, FILE *fp)
{
	int i, j;
	char *line;
	adr_rec_t *rec;

	rec = NULL;
	while((line = read_line(fp))){
		if(!line[0]){
			if(rec){
				if(rec->data[ADR_NAME].buf){
					adr_rec_add(db, rec);
					if(rec->priv)
						db->num_priv++;
					db->num_recs++;
				}
				rec = NULL;
			}
		}else
		if(!strcmp(line, "private: yes")){
			if(!rec)
				rec = (adr_rec_t *)zalloc(sizeof(adr_rec_t));
			rec->priv = 1;
		}else
		for(i = 0; i < ADR_NUM_FIELDS; i++){
			j = strlen(adr_field[i]);
			if(strstr(line, adr_field[i]) != line ||
					strstr(line, ": ") != line + j)
				continue;
			if(!rec)
				rec = (adr_rec_t *)zalloc(sizeof(adr_rec_t));
			rec->data[i].size = strlen(line + j + 2);
			if(rec->data[i].buf)
				free(rec->data[i].buf);
			rec->data[i].buf = (u8 *)malloc(rec->data[i].size + 1);
			strcpy(rec->data[i].buf, line + j + 2);
			sub_chr(rec->data[i].buf, '\t', '\n');
		}
		free(line);
	}

	return;
}

void
adr_print(adr_t *db, FILE *fp)
{
	int i;
	adr_rec_t *rec;

	for(rec = db->head; rec; rec = rec->next){
		for(i = 0; i < ADR_NUM_FIELDS; i++){
			if(rec->data[i].buf){
				fprintf(fp, "%s: ", adr_field[i]);
				print_str(fp, rec->data[i].buf);
				fprintf(fp, "\n");
			}
		}
		if(rec->priv)
			fprintf(fp, "private: yes\n");
		fprintf(fp, "\n");
	}

	return;
}

void
adr_init(adr_t *db)
{
	db->num_priv = db->num_recs = 0;
	db->head = db->tail = NULL;

	return;
}

void
adr_free(adr_t *db)
{
	adr_rec_t *rec, *rec_next;

	for(rec = db->head; rec; rec = rec_next){
		rec_next = rec->next;
		adr_rec_free(rec);
	}

	return;
}

void
adr_rec_free(adr_rec_t *rec)
{
	int i;

	for(i = 0; i < ADR_NUM_FIELDS; i++){
		if(rec->data[i].buf)
			free(rec->data[i].buf);
	}
	if(rec->mmp)
		mmp_free(rec->mmp);
	free(rec);

	return;
}

DEFINE_ADD(adr, adr_rec, )

DEFINE_DEL(adr, adr_rec, )

DEFINE_QSORT(adr_rec)

void
adr_rec_sort(adr_t *db)
{
	adr_rec_qsort(db->head, db->tail);

	return;
}

int
adr_rec_cmp(adr_rec_t *a, adr_rec_t *b)
{
	int a0, b0;

	a0 = isalpha(a->data[ADR_NAME].buf[0]);
	b0 = isalpha(b->data[ADR_NAME].buf[0]);

	if((a0 && b0) || (!a0 && !b0))
		return strcmp(a->data[ADR_NAME].buf, b->data[ADR_NAME].buf);

	if(a0)
		return -1;

	return 1;
}
