/* $Id: define.h,v 1.5 2005/11/28 09:21:47 geni Exp $
 *
 * Copyright (c) 2005 Huidae Cho
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef _DEFINE_H_
#define _DEFINE_H_

#define DEFINE_ADD(pim, type, prefix) \
void \
type##prefix##_add(pim##_t *db, type##_t *rec) \
{ \
	type##_t *r; \
 \
	if(!rec) \
		return; \
 \
	for(r = db->prefix##head; r && type##_cmp(rec, r) >= 0; r = r->next); \
	if(r){ \
		rec->prev = r->prev; \
		if(rec->prev) \
			rec->prev->next = rec; \
		else \
			db->prefix##head = rec; \
		rec->next = r; \
		rec->next->prev = rec; \
	}else{ \
		rec->prev = db->prefix##tail; \
		if(rec->prev) \
			rec->prev->next = rec; \
		else \
			db->prefix##head = rec; \
		rec->next = NULL; \
		db->prefix##tail = rec; \
	} \
 \
	return; \
}

#define DEFINE_DEL(pim, type, prefix) \
void \
type##prefix##_del(pim##_t *db, type##_t *rec) \
{ \
	if(!rec) \
		return; \
 \
	if(rec->prev) \
		rec->prev->next = rec->next; \
	if(rec->next) \
		rec->next->prev = rec->prev; \
	if(rec == db->prefix##head) \
		db->prefix##head = rec->next; \
	if(rec == db->prefix##tail) \
		db->prefix##tail = rec->prev; \
 \
	return; \
}

/* general quick sort definition */
#define DEFINE_QSORT(type) \
static void \
type##_qsort(type##_t *head, type##_t *tail) \
{ \
	int l; \
	type##_t pivot, *p, *h, *t; \
 \
	if(!head || !tail) \
		return; \
 \
	l = sizeof(type##_t) - 2 * sizeof(type##_t *); \
 \
	h = head; \
	t = tail; \
	memcpy(&pivot, head, l); \
 \
	while(head != tail){ \
		while(type##_cmp(tail, &pivot) >= 0 && head != tail) \
			tail = tail->prev; \
		if(head != tail){ \
			memcpy(head, tail, l); \
			head = head->next; \
		} \
		while(type##_cmp(head, &pivot) <= 0 && head != tail) \
			head = head->next; \
		if(head != tail){ \
			memcpy(tail, head, l); \
			tail = tail->prev; \
		} \
	} \
 \
	memcpy(head, &pivot, l); \
	p = head; \
	head = h; \
	tail = t; \
	if(head != p) \
		type##_qsort(head, p->prev); \
	if(tail != p) \
		type##_qsort(p->next, tail); \
 \
	return; \
}

#define _open() \
	do{if((fd = open(file, O_WRITE, S_IRWUSR)) < 0){ \
		close(fd); \
		ERROR("%s: open failed", file); \
		return -1; \
	}}while(0)
#define _write(data, size) \
	do{if(write(fd, data, size) != size){ \
		close(fd); \
		ERROR("%s: write failed", file); \
		return -1; \
	}}while(0)
#define _close() \
	close(fd)

#endif
