/* $Id: mmp.c,v 1.6 2005/11/27 06:23:43 geni Exp $
 *
 * Copyright (c) 2005 Huidae Cho
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include "onhand.h"
#include "define.h"

int
mmp_create(mmp_t *mmp, u16 width, u16 height)
{
	/* height should be a multiple of 8 */
	height = (height + 7) / 8 * 8;

	mmp->width = width;
	mmp->height = height;

	if(width && height)
		mmp->buf = (u8 *)zalloc(width * height / 8);
	else
		mmp->buf = NULL;

	return 0;
}

int
mmp_create_from_pbm(mmp_t *mmp, char *file)
{
	int fd, i, j, k;
	u8 i8;
	u16 i16, w, h;

	if((fd = open(file, O_RDONLY)) < 0){
		ERROR("%s: open failed", file);
		return -1;
	}

	if(read(fd, &i16, 2) != 2){
		close(fd);
		ERROR("%s: read failed", file);
		return -1;
	}

	if(read(fd, &i8, 1) != 1){
		close(fd);
		ERROR("%s: read failed", file);
		return -1;
	}

	if(i16 != 0x3450 ||
			(i8 != 0x20 && i8 != 0x09 && i8 != 0x0d && i8 != 0x0a)){
		close(fd);
		ERROR("%s: not a pbm file", file);
		return -1;
	}

	w = 0;
	/* skip comment, if any */
	if(read(fd, &i8, 1) != 1){
		close(fd);
		ERROR("%s: read failed", file);
		return -1;
	}
	if(i8 == '#')
		while(read(fd, &i8, 1) == 1 && i8);
	else
		w = i8 - '0';

	for(; read(fd, &i8, 1) == 1 && i8 >= '0' && i8 <= '9'; )
		w = 10 * w + i8 - '0';

	for(h = 0; read(fd, &i8, 1) == 1 && i8 >= '0' && i8 <= '9'; )
		h = 10 * h + i8 - '0';

	if(w < 1 || w > MAX_WIDTH || h < 1 || h > MAX_HEIGHT){
		close(fd);
		ERROR("%d*%d: invalid image size", w, h);
		return -1;
	}

	mmp_create(mmp, w, h);

	for(i = k = 0; read(fd, &i8, 1) == 1; i += 8){
		for(j = 0; j < 8 && i + j < w; j++){
			if(i8 & (1 << (7-j)))
				mmp_set(mmp, i + j, k);
		}
		if(i + j == w){
			i = -8;
			k++;
		}
	}

	return 0;
}

void
mmp_free(mmp_t *mmp)
{
	if(!mmp)
		return;

	if(mmp->buf)
		free(mmp->buf);

	return;
}

int
mmp_write(mmp_t *mmp, char *file)
{
	int fd, i;

	_open();

	_write(&mmp->width, 2);
	_write(&mmp->height, 2);
	i = mmp->width * mmp->height / 8;
	_write(mmp->buf, i);

	_close();

	return 0;
}
