/* $Id: mmp_mkfont.c,v 1.6 2005/11/21 06:25:50 geni Exp $
 *
 * Copyright (c) 2005 Huidae Cho
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include "onhand.h"

int
main(int argc, char *argv[])
{
	FILE *fp;
	mmp_t m;
	int x, y;

#ifdef TINYHAN8
	fp = fopen("tinyhan8.txt", "r");
	mmp_create(&m, 304, 496);
	for(y = 0; y < m.height; y++){
		for(x = 0; x < m.width; x++){
			if(fgetc(fp) == '#')
				mmp_set(&m, x, y);
		}
		fgetc(fp);
	}
	fclose(fp);
#else
	int x1, y1, x2, y2;

	fp = fopen("tinyhan9.txt", "r");
	mmp_create(&m, 432, 392);
	for(y = 0; y < 558; y++){
		for(x = 0; x < 304; x++){
			if(fgetc(fp) == '#'){
				/* serialize and rearrange:
				 * 304*558  8*21204  432*392 (mmp font)
				 * 8*9      8*9      8*8 (unit)
				 * 38*62    1*2356   54*49 (dimension)
				 * 38*62    1*2356   48*49 (characters)
				 * +-----+    +-+    +------+
				 * |     | => | | => |      |
				 * |     |    | |    |      |
				 * +-----+    | |    +------+
				 *            | |
				 *            +-+
				 * index = 0 ... 2349
				 * font byte = (9 * index) / 8 * 8
				 * font bit = index % 8
				 *
				 * font bit 7 = (index = 47) % 8
				 * font byte = (9 * 47) / 8 * 8 = 416
				 * width = 416 + 8 * 2 = 432
				 * height = ceil(2350 / (47 + 1)) * 8 = 392
				 */
				x1 = x % 8;
				y1 = ((y / 9) * 38 + x / 8) * 9 + y % 9;
				x2 = ((y1 / 8) % 54) * 8 + x1 % 8;
				y2 = ((y1 / 8) / 54) * 8 + y1 % 8;
				mmp_set(&m, x2, y2);
			}
		}
		fgetc(fp);
	}
	fclose(fp);
#endif
	mmp_write(&m, "tinyhan.mmp");
	mmp_free(&m);

	return 0;
}
