/* $Id: onhand.h,v 1.26 2005/11/29 19:29:04 geni Exp $
 *
 * Copyright (c) 2005 Huidae Cho
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef _ONHAND_H_
#define _ONHAND_H_

#include <stdio.h>

#define VERSION "0.0.1"

#define ERROR(fmt, args...) \
	fprintf(stderr, "%s (%d in %s): "fmt"\n", __FUNCTION__, __LINE__, \
			__FILE__, ##args)
#ifdef DEBUG
#define LINE ERROR("DEBUG");
#else
#define LINE
#endif

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;

#define zalloc(x) calloc(1, (x))

#define O_WRITE (O_CREAT | O_TRUNC | O_WRONLY)
#define S_IRWUSR (S_IRUSR | S_IWUSR)

/******************************************************************************/

/* mmp */

/* 0: width [2]
 * 2: height [2]
 * 4: buf [width * height / 8]
 */

#define MAX_WIDTH 320
#define MAX_HEIGHT 240

#define LCD_WIDTH 102
#define LCD_HEIGHT 64

typedef struct _mmp {
	u16 width,
	    height;	/* a multiple of 8 */
	u8 *buf;
} mmp_t;

typedef struct _mmv_frame {
	u16 size,
	    unknown1,
	    unknown2,
	    unknown3,
	    x,
	    y,	/* a multiple of 8 */
	    unknown4,
	    unknown5;
	mmp_t mmp;
} __attribute__((packed)) mmv__frame_t;

/* mmp.c */
int mmp_create(mmp_t *mmp, u16 width, u16 height);
int mmp_create_from_pbm(mmp_t *mmp, char *file);
void mmp_free(mmp_t *mmp);
int mmp_write(mmp_t *mmp, char *file);
#define mmp_get(m, x, y) (((m)->buf[(y)/8*(m)->width+(x)] & (1<<((y)%8))) ? 1:0)
#define mmp_set(m, x, y) ((m)->buf[(y)/8*(m)->width+(x)] |= (1<<((y)%8)))
#define mmp_clr(m, x, y) ((m)->buf[(y)/8*(m)->width+(x)] &= ~(1<<((y)%8)))

/******************************************************************************/

#ifdef _LIBONHAND_MISC_C_
char *pim_magic = "K66E"; /* 4b 36 36 45 */
char *dow[] = {
	"sun",
	"mon",
	"tue",
	"wed",
	"thu",
	"fri",
	"sat",
};
#else
extern char *pim_magic, *dow[];
#endif

typedef struct _data {
	u16 size;
	u8 *buf;
} data_t;

typedef struct _date {
	u16 year;
	u8 month, day;
} date_t;

typedef struct _minute {
	u16 year, minute;
	u8 month, day;
} minute_t;

/******************************************************************************/

/* adr */

/* HEADER [224]:
 * 0: magic [4]: 4b 36 36 45
 * 4: num_priv [2]
 * 6: num_recs [2]: total available records excluding overwritten data
 * 8: offset [27*2*4]: (27: A-Z, other) * (first, last) * 4 byte offset
 * 	ff ff ff ff: no data
 * 	last == first: one data
 *
 * 224: RECORD [>=20]:
 * 0: prev [4] for A-Z, other:
 *  	ff ff ff ff: no data
 * 4: next [4] for A-Z, other:
 *  	ff ff ff ff: no data
 * 8: unknown [6]: => 00 00 00 00 5e cb
 * 14: private [1]
 * 15: unknown [1]: => 00
 * 16: image [1]
 * 17: unknown [1]: => 00
 * 18: data size [2]
 * 20: 13 * {
 * 	size [2]
 * 	data [size]
 * }
 * if(image){
 * 	width [2]
 * 	height [2]
 * 	buf [width * height / 8]
 * }
 */

typedef struct _adr {
	u16 num_priv, num_recs;
	struct _adr_rec *head, *tail;
} adr_t;

#define ADR_NAME 0
#define ADR_PHONE 1
#define ADR_TITLE 2
#define ADR_COMPANY 3
#define ADR_PHONE2 4
#define ADR_FAX 5
#define ADR_CELLULAR 6
#define ADR_PAGER 7
#define ADR_EMAIL 8
#define ADR_EMAIL2 9
#define ADR_UNKNOWN 10
#define ADR_ADDRESS 11
#define ADR_MEMO 12
#define ADR_NUM_FIELDS 13

#ifdef _LIBONHAND_MISC_C_
char *adr_field[] = {
	"name",
	"phone",
	"title",
	"company",
	"phone2",
	"fax",
	"cellular",
	"pager",
	"email",
	"email2",
	"unknown",
	"address",
	"memo",
};
#else
extern char *adr_field[];
#endif

typedef struct _adr_rec {
	data_t data[ADR_NUM_FIELDS];
	mmp_t *mmp;
	u8 priv;
	struct _adr_rec *prev, *next;
} adr_rec_t;

/* adr.c */
int adr_read(adr_t *db, char *file);
int adr_write(adr_t *db, char *file);
void adr_scan(adr_t *db, FILE *fp);
void adr_print(adr_t *db, FILE *fp);
void adr_init(adr_t *db);
void adr_free(adr_t *db);
void adr_rec_free(adr_rec_t *rec);
void adr_rec_add(adr_t *db, adr_rec_t *rec);
void adr_rec_del(adr_t *db, adr_rec_t *rec);
void adr_rec_sort(adr_t *db);
int adr_rec_cmp(adr_rec_t *a, adr_rec_t *b);

/******************************************************************************/

/* scl */

/* HEADER [12]:
 * 0: magic [4]: 4b 36 36 45
 * 4: num_priv [2]
 * 6: weekly holiday [2]:
 * 	bit: sun(0x1) mon(0x2) tue(0x4) wed(0x8) thu(0x10) fri(0x20) sat(0x40)
 * 8: offset for the first year [4]
 *
 * *: YEAR HEADER [106]:
 * 0: year [2]
 * 2: offset for next year [4]:
 * 	ff ff ff ff: no data
 * 6: offset for the first month of the year [4]:
 * 	ff ff ff ff: no data
 * 10: holiday [12*4]: (12 months) * 4 byte data
 * 	1 << (day-1)
 * 58: workday [12*4]: (12 months) * 4 byte data
 * 	1 << (day-1)
 *
 * *: MONTH HEADER [134]:
 * 0: month [1]
 * 1: unknown [1]
 * 2: offset for next month [4]: (December has ff ff ff ff)
 * 6: offset for daily records [31*4]: (31 days) * 4 byte offset
 * 130: offset for cross day records [4]:
 * 	If the record is cross-month, there are two exactly same records in
 * 	both months. The maximum length of a cross-month record is one month.
 *
 * *: RECORD [>=46]:
 * 0: prev [4]:
 *  	ff ff ff ff: no data
 * 4: next [4]:
 *  	ff ff ff ff: no data
 * 8: start day [1]
 * 9: start month [1]
 * 10: start year [2]
 * 12: end day [1]
 * 13: end month [1]
 * 14: end year [2]
 * 16: start minute [2]
 * 18: end minute [2]
 * 20: unknown [12]: => 2 * {00 00 ?? ?? ?? cb}
 * 32: unknown [1]
 * 33: cross day [1]:
 * 	00: one day
 * 	80: cross day
 * 34: private [1]
 * 35: unknown [1]
 * 36: data size [2]: title, place, memo, ...
 * 38: 3 * {
 * 	size [2]
 * 	data [size]
 * }
 * *: alarm [1]
 * *+1: unknown [1]
 * if(alarm == 1){
 * 	*+2: melody [8]
 * 	*+10: before minute [2]
 * }
 */

typedef struct _scl {
	u16 num_priv, holiday;
	struct _scl_holiday *hhead, *htail, *whead, *wtail;
	struct _scl_rec *head, *tail;
} scl_t;

typedef struct _scl_holiday {
	u16 year;
	u8 month;
	u32 holiday;
	struct _scl_holiday *prev, *next;
} scl_holiday_t;

#define SCL_TITLE 0
#define SCL_PLACE 1
#define SCL_MEMO 2
#define SCL_NUM_FIELDS 3

#ifdef _LIBONHAND_MISC_C_
char *scl_field[] = {
	"title",
	"place",
	"memo",
};
#else
extern char *scl_field[];
#endif

typedef struct _scl_rec {
	data_t data[SCL_NUM_FIELDS];
	minute_t smin, emin;
	u8 melody[9];
	u16 before;
	u8 priv;
	struct _scl_rec *prev, *next;
} scl_rec_t;

/* scl.c */
int scl_read(scl_t *db, char *file);
int scl_write(scl_t *db, char *file);
void scl_scan(scl_t *db, FILE *fp);
void scl_print(scl_t *db, FILE *fp);
void scl_init(scl_t *db);
void scl_free(scl_t *db);
void scl_rec_free(scl_rec_t *rec);
void scl_rec_add(scl_t *db, scl_rec_t *rec);
void scl_rec_del(scl_t *db, scl_rec_t *rec);
void scl_rec_sort(scl_t *db);
int scl_rec_cmp(scl_rec_t *a, scl_rec_t *b);
void scl_holidayh_add(scl_t *db, scl_holiday_t *holiday);
void scl_holidayh_del(scl_t *db, scl_holiday_t *holiday);
void scl_holidayw_add(scl_t *db, scl_holiday_t *holiday);
void scl_holidayw_del(scl_t *db, scl_holiday_t *holiday);
void scl_holidayh_sort(scl_t *db);
void scl_holidayw_sort(scl_t *db);
int scl_holiday_cmp(scl_holiday_t *a, scl_holiday_t *b);

/******************************************************************************/

/* tdo */

/* HEADER [16]:
 * 0: magic [4]: 4b 36 36 45
 * 4: num_priv [2]
 * 6: num_recs [2]: total records including overwritten data
 * 8: offset [2*4]: (first, last) * 4 byte offset
 * 	ff ff ff ff: no data
 * 	last == first: one data
 *
 * 16: RECORD [>=24]:
 * 0: prev [4]:
 *  	ff ff ff ff: no data
 * 4: next [4]:
 *  	ff ff ff ff: no data
 * 8: unknown [6]: => 00 00 00 00 68 cb
 * 14: private [1]
 * 15: unknown [1]: => 00
 * 16: data size [2]: size of title and memo (2 * {..})
 * 18: status [1]:
 * 	0: in progress
 * 	1: completed
 * 19: priority [1]:
 * 	0: none
 * 	1: low
 * 	2: normal
 * 	3: high
 * 20: day [1]
 * 21: month [1]
 * 22: year [2]
 * 24: 2 * {
 * 	size [2]
 * 	data [size]
 * }
 */

typedef struct _tdo {
	u16 num_priv, num_recs;
	struct _tdo_rec *head, *tail;
} tdo_t;

#define TDO_TITLE 0
#define TDO_MEMO 1
#define TDO_NUM_FIELDS 2

#ifdef _LIBONHAND_MISC_C_
char *tdo_field[] = {
	"title",
	"memo",
};
#else
extern char *tdo_field[];
#endif

typedef struct _tdo_rec {
	data_t data[TDO_NUM_FIELDS];
	date_t due;
	u8 status, priority, priv;
	struct _tdo_rec *prev, *next;
} tdo_rec_t;

/* tdo.c */
int tdo_read(tdo_t *db, char *file);
int tdo_write(tdo_t *db, char *file);
void tdo_scan(tdo_t *db, FILE *fp);
void tdo_print(tdo_t *db, FILE *fp);
void tdo_init(tdo_t *db);
void tdo_free(tdo_t *db);
void tdo_rec_free(tdo_rec_t *rec);
void tdo_rec_add(tdo_t *db, tdo_rec_t *rec);
void tdo_rec_del(tdo_t *db, tdo_rec_t *rec);
void tdo_rec_sort(tdo_t *db);
int tdo_rec_cmp(tdo_rec_t *a, tdo_rec_t *b);

/******************************************************************************/

/* txm */

/* HEADER [16]:
 * 0: magic [4]: 4b 36 36 45
 * 4: num_priv [2]
 * 6: num_recs [2]: total records including overwritten data
 * 8: offset [2*4]: (first, last) * 4 byte offset
 * 	ff ff ff ff: no data
 * 	last == first: one data
 *
 * 16: RECORD [>=18]:
 * 0: prev [4]:
 *  	ff ff ff ff: no data
 * 4: next [4]:
 *  	ff ff ff ff: no data
 * 8: unknown [6]: => 00 00 00 00 6e cb
 * 14: private [1]
 * 15: unknown [1]: => 00
 * 16: data size [2]
 * 18: 2 * {
 * 	size [2]
 * 	data [size]
 * }
 */

typedef struct _txm {
	u16 num_priv, num_recs;
	struct _txm_rec *head, *tail;
} txm_t;

#define TXM_TITLE 0
#define TXM_MEMO 1
#define TXM_NUM_FIELDS 2

#ifdef _LIBONHAND_MISC_C_
char *txm_field[] = {
	"title",
	"memo",
};
#else
extern char *txm_field[];
#endif

typedef struct _txm_rec {
	data_t data[TXM_NUM_FIELDS];
	u8 priv;
	struct _txm_rec *prev, *next;
} txm_rec_t;

/* txm.c */
int txm_read(txm_t *db, char *file);
int txm_write(txm_t *db, char *file);
void txm_scan(txm_t *db, FILE *fp);
void txm_print(txm_t *db, FILE *fp);
void txm_init(txm_t *db);
void txm_free(txm_t *db);
void txm_rec_free(txm_rec_t *rec);
void txm_rec_add(txm_t *db, txm_rec_t *rec);
void txm_rec_del(txm_t *db, txm_rec_t *rec);
void txm_rec_sort(txm_t *db);
int txm_rec_cmp(txm_rec_t *a, txm_rec_t *b);

/******************************************************************************/

/* xps */

/* HEADER [32]:
 * 0: magic [4]: 4b 36 36 45
 * 4: unknown [4]: => 00 00 00 00
 * 8: default group [2]:
 * 	ff ff: all groups
 * 	system/ini/expsbjct.ini default code
 * 	65 00: business
 * 	66 00: private
 * 	67 00: other
 * 	68 00: group 4
 * 	69 00: group 5
 * 	6a 00: group 6
 * 	6b 00: group 7
 * 10: default currency [2]:
 * 	system/ini/expsbjct.ini default code
 * 	c9 00: $
 * 	ca 00: EUR
 * 	cb 00: POUND
 * 	cc 00: DM
 * 	cd 00: F
 * 	ce 00: SFr
 * 	cf 00: YEN
 * 12: unknown [2]: => 01 00
 * 14: beginning display [2]:
 * 	00 00: monthly list
 * 	01 00: daily list
 * 	02 00: input
 * 16: unknown [12]: 02 00 00 ... 00
 * 28: offset for the first year [4]
 *
 * *: YEAR HEADER [10]:
 * 0: year [2]
 * 2: offset for next year [4]:
 * 	ff ff ff ff: no data
 * 6: offset for the first month of the year [4]:
 * 	ff ff ff ff: no data
 *
 * *: MONTH HEADER [998]:
 * 0: month [1]
 * 1: unknown [1]
 * 2: offset for next month [4]: (December has ff ff ff ff)
 * 6: daily total [7*31*4]: (7 groups) * (31 days) * 4 byte total
 * 874: offset for daily records [31*4]: (31 days) * 4 byte offset
 *
 * *: RECORD [64]:
 * 0: prev [4]:
 *  	ff ff ff ff: no data
 * 4: next [4]:
 *  	ff ff ff ff: no data
 * 8: overwritten [1]
 * 9: unknown [1]
 * 10: unknown [6]: => 00 00 ?? ?? ?? cb
 * 16: day [1]
 * 17: month [1]
 * 18: year [2]
 * 20: currency code [2]
 * 22: expense [4]: dollar (in cents: dollars * 100 + cents)
 * 26: unknown [2]: 00 00
 * 28: group code [2]
 * 30: memo [34]
 */

typedef struct _xps {
	u16 group, currency, display;
	struct _xps_rec *head, *tail;
} xps_t;

typedef struct _xps_rec {
	date_t date;
	u16 currency, group;
	u32 expense;
	u8 memo[35];
	struct _xps_rec *prev, *next;
} xps_rec_t;

/* xps.c */
int xps_read(xps_t *db, char *file);
int xps_write(xps_t *db, char *file);
void xps_scan(xps_t *db, FILE *fp);
void xps_print(xps_t *db, FILE *fp);
void xps_init(xps_t *db);
void xps_free(xps_t *db);
void xps_rec_free(xps_rec_t *rec);
void xps_rec_add(xps_t *db, xps_rec_t *rec);
void xps_rec_del(xps_t *db, xps_rec_t *rec);
void xps_rec_sort(xps_t *db);
int xps_rec_cmp(xps_rec_t *a, xps_rec_t *b);

/******************************************************************************/

/* misc.c */
char *squeeze_spaces(char *buf);
char *read_line(FILE *fp);
int read_file(char *file, u8 **data, int *size);
int str_chr(char *str, char ch);
int str_str(char *big, char *litttle);
void sub_chr(char *str, char a, char b);
void print_str(FILE *fp, char *str);
int day_of_week(int y, int m, int d);

#endif
