/* $Id: scl.c,v 1.15 2006/08/07 15:55:51 geni Exp $
 *
 * Copyright (c) 2005 Huidae Cho
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "onhand.h"
#include "define.h"

static int scl_rec_cmp_write_mode = 0;

static void scl_rec_qsort(scl_rec_t *head, scl_rec_t *tail);
static void scl_holiday_qsort(scl_holiday_t *head, scl_holiday_t *tail);

int
scl_read(scl_t *db, char *file)
{
	int size, i, j;
	u16 year;
	u8 *data, *ptr, *ptr2, month;
	u32 offset, offset2, offset3, holiday;
	scl_holiday_t *h;
	scl_rec_t *rec;

	if(read_file(file, &data, &size)){
		ERROR("read_file");
		return -1;
	}
	/* the minimum size is 12 = 4+2+2+4 */
	if(size < 12 || memcmp(data, pim_magic, 4)){
		ERROR("%s: not an scl file", file);
		return -1;
	}

	scl_init(db);

	memcpy(&db->holiday, data+6, 2);

	memcpy(&offset, data+8, 4);
	while(offset != 0xffffffff && offset < size){
		ptr = data + offset;
		memcpy(&year, ptr, 2);
		memcpy(&offset, ptr+2, 4);
		memcpy(&offset2, ptr+6, 4);
		/* holiday */
		for(i = 0; i < 12; i++){
			memcpy(&holiday, ptr+10+4*i, 4);
			if(!holiday)
				continue;
			h = (scl_holiday_t *)malloc(sizeof(scl_holiday_t));
			h->year = year;
			h->month = i + 1;
			h->holiday = holiday;
			h->prev = h->next = NULL;
			if(db->hhead){
				h->prev = db->htail;
				db->htail->next = h;
				db->htail = h;
			}else
				db->hhead = db->htail = h;
		}
		/* workday */
		for(i = 0; i < 12; i++){
			memcpy(&holiday, ptr+58+4*i, 4);
			if(!holiday)
				continue;
			h = (scl_holiday_t *)malloc(sizeof(scl_holiday_t));
			h->year = year;
			h->month = i + 1;
			h->holiday = holiday;
			h->prev = h->next = NULL;
			if(db->whead){
				h->prev = db->wtail;
				db->wtail->next = h;
				db->wtail = h;
			}else
				db->whead = db->wtail = h;
		}
		while(offset2 != 0xffffffff && offset2 < size){
			ptr = data + offset2;
			memcpy(&month, ptr, 1);
			memcpy(&offset2, ptr+2, 4);
			for(i = 0; i < 32; i++){
				memcpy(&offset3, ptr+6+4*i, 4);
				while(offset3 != 0xffffffff && offset3 < size){
					ptr2 = data + offset3;
					memcpy(&offset3, ptr2+4, 4);
					if(ptr2[33] == 0x80 && month != ptr2[9])
						continue;
					rec = (scl_rec_t *)zalloc(
							sizeof(scl_rec_t));
					rec->smin.day = ptr2[8];
					rec->smin.month = ptr2[9];
					memcpy(&rec->smin.year, ptr2+10, 2);
					rec->emin.day = ptr2[12];
					rec->emin.month = ptr2[13];
					memcpy(&rec->emin.year, ptr2+14, 2);
					memcpy(&rec->smin.minute, ptr2+16, 2);
					memcpy(&rec->emin.minute, ptr2+18, 2);
					rec->priv = ptr2[34];
					if(rec->priv)
						db->num_priv++;
					ptr2 += 38;
					for(j = 0; j < SCL_NUM_FIELDS; j++){
						memcpy(&rec->data[j].size, ptr2,
								2);
						ptr2 += 2;
						rec->data[j].buf = NULL;
						if(!rec->data[j].size)
							continue;
						rec->data[j].buf = (u8 *)malloc(
							rec->data[j].size+1);
						memcpy(rec->data[j].buf, ptr2,
							rec->data[j].size);
						rec->data[j].buf[
							rec->data[j].size] = 0;
						ptr2 += rec->data[j].size;
					}
					if(*ptr2){
						ptr2 += 2;
						memcpy(rec->melody, ptr2, 8);
						rec->melody[8] = 0;
						ptr2 += 8;
						memcpy(&rec->before, ptr2, 2);
					}
					rec->prev = rec->next = NULL;
					if(db->head){
						rec->prev = db->tail;
						db->tail->next = rec;
						db->tail = rec;
					}else
						db->head = db->tail = rec;
				}
			}
		}
	}

	free(data);

	scl_rec_sort(db);

	return 0;
}

int
scl_write(scl_t *db, char *file)
{
	int fd, i, j;
	u8 i8;
	u16 i16;
	u32 i32, offset, offset2, offset3 = 0;
	date_t date;
	scl_holiday_t *h, *w;
	scl_rec_t *rec, *r;

	_open();

	_write(pim_magic, 4);
	_write(&db->num_priv, 2);
	_write(&db->holiday, 2);

	/* header */
	if(db->head)
		offset = 12;
	else
		offset = 0xffffffff;

	_write(&offset, 4);

	for(rec = db->head; rec; rec = rec->next){
		if(rec->smin.month == rec->emin.month ||
		   rec->smin.year > rec->emin.year ||
		   (rec->smin.year == rec->emin.year &&
		    rec->smin.month > rec->emin.month))
			continue;
		r = (scl_rec_t *)zalloc(sizeof(scl_rec_t));
		r->emin.year = rec->smin.year;
		r->emin.month = rec->smin.month;
		r->emin.day = rec->smin.day;
		r->emin.minute = rec->smin.minute;
		r->smin.year = rec->emin.year;
		r->smin.month = rec->emin.month;
		r->smin.day = rec->emin.day;
		r->smin.minute = rec->emin.minute;
		r->before = rec->before;
		r->priv = rec->priv;
		memcpy(r->melody, rec->melody, 9);
		for(i = 0; i < SCL_NUM_FIELDS; i++){
			r->data[i].size = rec->data[i].size;
			if(!r->data[i].size)
				continue;
			r->data[i].buf = (u8 *)malloc(sizeof(r->data[i].size));
			memcpy(r->data[i].buf, rec->data[i].buf,
					r->data[i].size);
		}
		scl_rec_add(db, r);
	}

	scl_rec_cmp_write_mode = 1;
	scl_rec_sort(db);
	scl_rec_cmp_write_mode = 0;

	h = db->hhead;
	w = db->whead;

	date.year = 0;
	for(rec = db->head; rec; rec = rec->next){
		if(date.year != rec->smin.year){
			date.year = rec->smin.year;
			/* year header */
			/* year */
			_write(&date.year, 2);

			offset += 106;
			offset2 = offset;
			date.month = 0;
			for(r = rec; r && date.year == r->smin.year;
					r = r->next){
				if(date.month != r->smin.month){
					date.month = r->smin.month;
					offset2 += 134;
				}
				offset2 += 46;
				for(i = 0; i < SCL_NUM_FIELDS; i++)
					offset2 += r->data[i].size;
				if(r->melody[0])
					offset2 += 10;
			}
			if(!r)
				offset2 = 0xffffffff;
			/* offset for next year */
			_write(&offset2, 4);

			/* offset for the first month of the year */
			_write(&offset, 4);

			/* holiday */
			for(i = 1; i < 13; i++){
				i32 = 0;
				for(; h && date.year > h->year; h = h->next);
				if(h && date.year == h->year){
					for(; h && i > h->month; h = h->next);
					if(h && i == h->month)
						i32 = h->holiday;
				}
				_write(&i32, 4);
			}
			/* workday */
			for(i = 1; i < 13; i++){
				i32 = 0;
				for(; w && date.year > w->year; w = w->next);
				if(w && date.year == w->year){
					for(; w && i > w->month; w = w->next);
					if(w && i == w->month)
						i32 = w->holiday;
				}
				_write(&i32, 4);
			}

			date.month = 0;
		}
		if(date.month != rec->smin.month){
			/* month header */
			/* month */
			date.month = rec->smin.month;
			_write(&date.month, 1);
			i8 = 0;
			_write(&i8, 1);

			offset += 134;
			offset2 = offset;
			if(date.month < 12){
				for(r = rec; r && date.year == r->smin.year &&
						date.month == r->smin.month;
						r = r->next){
					offset2 += 46;
					for(i = 0; i < SCL_NUM_FIELDS; i++)
						offset2 += r->data[i].size;
					if(r->melody[0])
						offset2 += 10;
				}
				if(!r || date.year != r->smin.year)
					offset2 = 0xffffffff;
			}else
				offset2 = 0xffffffff;
			/* offset for next month */
			_write(&offset2, 4);

			/* offset for daily records */
			offset2 = offset;
			r = rec;
			for(i = 1; i < 32; i++){
				for(; r && date.year == r->smin.year &&
					date.month == r->smin.month &&
					date.year == r->emin.year &&
					date.month == r->emin.month &&
					i > r->smin.day; r = r->next){
					offset2 += 46;
					for(j = 0; j < SCL_NUM_FIELDS; j++)
						offset2 += r->data[j].size;
					if(r->melody[0])
						offset2 += 10;
				}
				if(r && date.year == r->smin.year &&
					date.month == r->smin.month &&
					date.year == r->emin.year &&
					date.month == r->emin.month &&
					i == r->smin.day)
					i32 = offset2;
				else
					i32 = 0xffffffff;
				_write(&i32, 4);
			}

			/* cross-month offset */
			for(; r && date.year == r->smin.year &&
				date.month == r->smin.month &&
				date.year == r->emin.year &&
				date.month == r->emin.month &&
				31 == r->smin.day; r = r->next){
				offset2 += 46;
				for(j = 0; j < SCL_NUM_FIELDS; j++)
					offset2 += r->data[j].size;
				if(r->melody[0])
					offset2 += 10;
			}
			if(r && r->smin.month != r->emin.month &&
					date.year == r->smin.year &&
					date.month == r->smin.month)
				i32 = offset2;
			else
				i32 = 0xffffffff;
			_write(&i32, 4);

			date.day = 0;
		}

		/* prev */
		i32 = offset3;
		if(rec->smin.month == rec->emin.month){
			if(date.day != rec->smin.day){
				date.day = rec->smin.day;
				i32 = 0xffffffff;
			}
		}else{
			if(!rec->prev ||
			   rec->prev->smin.month == rec->prev->emin.month ||
			   rec->prev->smin.month != rec->smin.month)
				i32 = 0xffffffff;
		}
		_write(&i32, 4);

		offset3 = offset;

		if(rec->next){
			offset += 46;
			for(j = 0; j < SCL_NUM_FIELDS; j++)
				offset += rec->data[j].size;
			if(rec->melody[0])
				offset += 10;
		}

		/* next */
		i32 = offset;
		if(!rec->next || rec->smin.month != rec->next->smin.month)
			i32 = 0xffffffff;
		else
		if(rec->smin.month == rec->emin.month){
			if(rec->next->smin.month != rec->next->emin.month ||
					rec->smin.day != rec->next->smin.day)
				i32 = 0xffffffff;
		}
		_write(&i32, 4);

		/* data */
		if(rec->smin.year > rec->emin.year ||
				(rec->smin.year == rec->emin.year &&
				 rec->smin.month > rec->emin.month)){
			_write(&rec->emin.day, 1);
			_write(&rec->emin.month, 1);
			_write(&rec->emin.year, 2);
			_write(&rec->smin.day, 1);
			_write(&rec->smin.month, 1);
			_write(&rec->smin.year, 2);
		}else{
			_write(&rec->smin.day, 1);
			_write(&rec->smin.month, 1);
			_write(&rec->smin.year, 2);
			_write(&rec->emin.day, 1);
			_write(&rec->emin.month, 1);
			_write(&rec->emin.year, 2);
		}
		_write(&rec->smin.minute, 2);
		_write(&rec->emin.minute, 2);

		i32 = 0;
		/* TODO: 00 ... 00 for unknown works! */
		for(i = 0; i < 3; i++)
			_write(&i32, 4);
		i8 = 0;
		_write(&i8, 1);

		if(rec->smin.year != rec->emin.year ||
				rec->smin.month != rec->emin.month)
			i8 = 0x80;
		else
			i8 = 0;
		_write(&i8, 1);
		_write(&rec->priv, 1);
		i8 = 0;
		_write(&i8, 1);

		i16 = 2;
		for(i = 0; i < SCL_NUM_FIELDS; i++)
			i16 += 2 + rec->data[i].size;
		if(rec->melody[0])
			i16 += 10;
		_write(&i16, 2);
		for(i = 0; i < SCL_NUM_FIELDS; i++){
			_write(&rec->data[i].size, 2);
			if(!rec->data[i].size)
				continue;
			_write(rec->data[i].buf, rec->data[i].size);
		}
		if(rec->melody[0])
			i8 = 1;
		else
			i8 = 0;
		_write(&i8, 1);
		i8 = 0;
		_write(&i8, 1);
		if(rec->melody[0]){
			_write(rec->melody, 8);
			_write(&rec->before, 2);
		}
	}

	for(rec = db->head; rec; rec = r){
		r = rec->next;
		if(rec->smin.month == rec->emin.month ||
				rec->smin.year < rec->emin.year ||
				(rec->smin.year == rec->emin.year &&
				 rec->smin.month < rec->emin.month))
			continue;
		scl_rec_del(db, rec);
		scl_rec_free(rec);
	}

	scl_rec_sort(db);

	_close();

	return 0;
}

void
scl_scan(scl_t *db, FILE *fp)
{
	int i, j;
	char *line, *ptr, *ptr2, tmp[9];
	u8 sh, sm, eh, em;
	date_t d;
	scl_holiday_t *h;
	scl_rec_t *rec;

	rec = NULL;
	while((line = read_line(fp))){
		if(!line[0]){
			if(rec){
				if(rec->data[SCL_TITLE].buf[0]){
					scl_rec_add(db, rec);
					if(rec->priv)
						db->num_priv++;
				}
				rec = NULL;
			}
		}else
		if(strstr(line, "weekly holiday: ") == line){
			ptr = line + 16;
			while(*ptr){
				for(; !isalpha(*ptr); ptr++);
				ptr2 = ptr;
				for(; isalpha(*ptr); ptr++);
				*ptr = 0;
				for(i = 0; i < 7 && strcmp(ptr2, dow[i]); i++);
				if(i < 7)
					db->holiday |= 1 << i;
				ptr++;
			}
		}else
		if(strstr(line, "holiday: ") == line){
			if(sscanf(line+9, "%04hd-%02hhd-%02hhd", &d.year,
						&d.month, &d.day) == 3){
			for(h = db->hhead; h; h = h->next){
				if(h->year == d.year && h->month == d.month)
					break;
			}
			if(!h){
				h = (scl_holiday_t *)zalloc(
						sizeof(scl_holiday_t));
				h->year = d.year;
				h->month = d.month;
				scl_holidayh_add(db, h);
			}
			h->holiday |= 1 << (d.day-1);
			}
		}else
		if(strstr(line, "workday: ") == line){
			if(sscanf(line+9, "%04hd-%02hhd-%02hhd", &d.year,
						&d.month, &d.day) == 3){
			for(h = db->whead; h; h = h->next){
				if(h->year == d.year && h->month == d.month)
					break;
			}
			if(!h){
				h = (scl_holiday_t *)zalloc(
						sizeof(scl_holiday_t));
				h->year = d.year;
				h->month = d.month;
				scl_holidayw_add(db, h);
			}
			h->holiday |= 1 << (d.day-1);
			}
		}else
		if(!strcmp(line, "private: yes")){
			if(!rec)
				rec = (scl_rec_t *)zalloc(sizeof(scl_rec_t));
			rec->priv = 1;
		}else
		if(strstr(line, "time: ") == line){
			if(sscanf(line+6, "%04hd-%02hhd-%02hhd", &d.year,
						&d.month, &d.day) == 3){
			if(!rec)
				rec = (scl_rec_t *)zalloc(sizeof(scl_rec_t));
			rec->smin.year = d.year;
			rec->smin.month = d.month;
			rec->smin.day = d.day;
			rec->smin.minute = rec->emin.minute = 0xffff;
			if(sscanf(line+16, " - %04hd-%02hhd-%02hhd",
					&rec->emin.year, &rec->emin.month,
					&rec->emin.day) != 3){
				rec->emin.year = rec->smin.year;
				rec->emin.month = rec->smin.month;
				rec->emin.day = rec->smin.day;
				if(sscanf(line+16,
					" %02hhd:%02hhd - %02hhd:%02hhd",
					&sh, &sm, &eh, &em) == 4){
					if(sh > 23)
						sh = 23;
					if(sm > 59)
						sm = 59;
					if(eh > 23)
						eh = 23;
					if(em > 59)
						em = 59;
					rec->smin.minute = sh * 60 + sm;
					rec->emin.minute = eh * 60 + em;
				}else
				if(sscanf(line+16, " %02hhd:%02hhd", &sh, &sm)
						== 2){
					if(sh > 23)
						sh = 23;
					if(sm > 59)
						sm = 59;
					rec->smin.minute = sh * 60 + sm;
					rec->emin.minute = rec->smin.minute;
				}
			}
			}
		}else
		if(strstr(line, "alarm: ") == line){
			memset(tmp, 0, 9);
			for(ptr = line + 7, i = 0; i < 8 && *ptr && *ptr != ' ';
					ptr++, i++)
				tmp[i] = toupper(*ptr);
			for(; *ptr != ' '; ptr++);
			if(i && sscanf(ptr, " before %02hhd:%02hhd", &sh, &sm)
					== 2){
			if(!rec)
				rec = (scl_rec_t *)zalloc(sizeof(scl_rec_t));
			memcpy(rec->melody, tmp, 9);
			rec->before = sh * 60 + sm;
			if(rec->before > 999)
				rec->before = 999;
			}
		}else
		for(i = 0; i < SCL_NUM_FIELDS; i++){
			j = strlen(scl_field[i]);
			if(strstr(line, scl_field[i]) != line ||
					strstr(line, ": ") != line + j)
				continue;
			if(!rec)
				rec = (scl_rec_t *)zalloc(sizeof(scl_rec_t));
			rec->data[i].size = strlen(line + j + 2);
			if(rec->data[i].buf)
				free(rec->data[i].buf);
			rec->data[i].buf = (u8 *)malloc(rec->data[i].size + 1);
			strcpy(rec->data[i].buf, line + j + 2);
			sub_chr(rec->data[i].buf, '\t', '\n');
		}
		free(line);
	}
	return;
}

void
scl_print(scl_t *db, FILE *fp)
{
	u32 i;
	scl_holiday_t *h;
	scl_rec_t *rec;

	if(db->holiday){
		fprintf(fp, "weekly holiday:");
		for(i = 0; i < 7; i++){
			if(db->holiday & (1<<i))
				fprintf(fp, " %s", dow[i]);
		}
		fprintf(fp, "\n");
	}
	for(h = db->hhead; h; h = h->next){
		for(i = 0; i < 31; i++){
			if(h->holiday & (1<<i))
				fprintf(fp, "holiday: %04d-%02d-%02d %s\n",
						h->year, h->month, i+1,
						dow[day_of_week(h->year,
							h->month, i+1)]);
		}
	}
	for(h = db->whead; h; h = h->next){
		for(i = 0; i < 31; i++){
			if(h->holiday & (1<<i))
				fprintf(fp, "workday: %04d-%02d-%02d %s\n",
						h->year, h->month, i+1,
						dow[day_of_week(h->year,
							h->month, i+1)]);
		}
	}
	fprintf(fp, "\n");
	for(rec = db->head; rec; rec = rec->next){
		fprintf(fp, "time: %04d-%02d-%02d", rec->smin.year,
				rec->smin.month, rec->smin.day);
		if(rec->smin.year != rec->emin.year ||
				rec->smin.month != rec->emin.month ||
				rec->smin.day != rec->emin.day)
			fprintf(fp, " - %04d-%02d-%02d", rec->emin.year,
				rec->emin.month, rec->emin.day);
		else
		if(rec->smin.minute != 0xffff &&
				rec->smin.minute == rec->emin.minute)
			fprintf(fp, " %02d:%02d",
				rec->smin.minute/60, rec->smin.minute%60);
		else
		if(rec->smin.minute != 0xffff && rec->emin.minute != 0xffff)
			fprintf(fp, " %02d:%02d - %02d:%02d",
				rec->smin.minute/60, rec->smin.minute%60,
				rec->emin.minute/60, rec->emin.minute%60);
		fprintf(fp, " %s", dow[day_of_week(rec->smin.year,
					rec->smin.month, rec->smin.day)]);
		if(rec->smin.year != rec->emin.year ||
				rec->smin.month != rec->emin.month ||
				rec->smin.day != rec->emin.day)
			fprintf(fp, " - %s", dow[day_of_week(rec->emin.year,
					rec->emin.month, rec->emin.day)]);
		fprintf(fp, "\n");
		for(i = 0; i < SCL_NUM_FIELDS; i++){
			if(rec->data[i].buf){
				fprintf(fp, "%s: ", scl_field[i]);
				print_str(fp, rec->data[i].buf);
				fprintf(fp, "\n");
			}
		}
		if(rec->melody[0])
			fprintf(fp, "alarm: %s before %02d:%02d\n", rec->melody,
					rec->before/60, rec->before%60);
		if(rec->priv)
			fprintf(fp, "private: yes\n");
		fprintf(fp, "\n");
	}

	return;
}

void
scl_init(scl_t *db)
{
	db->num_priv = db->holiday = 0;
	db->hhead = db->htail = db->whead = db->wtail = NULL;
	db->head = db->tail = NULL;

	return;
}

void
scl_free(scl_t *db)
{
	scl_rec_t *rec, *rec_next;
	scl_holiday_t *h, *h_next;

	for(rec = db->head; rec; rec = rec_next){
		rec_next = rec->next;
		scl_rec_free(rec);
	}
	for(h = db->hhead; h; h = h_next){
		h_next = h->next;
		free(h);
	}
	for(h = db->whead; h; h = h_next){
		h_next = h->next;
		free(h);
	}

	return;
}

void
scl_rec_free(scl_rec_t *rec)
{
	int i;

	for(i = 0; i < SCL_NUM_FIELDS; i++){
		if(rec->data[i].buf)
			free(rec->data[i].buf);
	}
	free(rec);

	return;
}

DEFINE_ADD(scl, scl_rec, )

DEFINE_DEL(scl, scl_rec, )

DEFINE_QSORT(scl_rec)

void
scl_rec_sort(scl_t *db)
{
	scl_rec_qsort(db->head, db->tail);

	return;
}

int
scl_rec_cmp(scl_rec_t *a, scl_rec_t *b)
{
	if(a->smin.year > b->smin.year)
		return 1;
	if(a->smin.year < b->smin.year)
		return -1;

	if(a->smin.month > b->smin.month)
		return 1;
	if(a->smin.month < b->smin.month)
		return -1;

	if(scl_rec_cmp_write_mode){
		if(a->smin.month != a->emin.month &&
				b->smin.month == b->emin.month)
			return 1;
		if(a->smin.month == a->emin.month &&
				b->smin.month != b->emin.month)
			return -1;
		if(a->smin.month != a->emin.month &&
				b->smin.month != b->emin.month){
			if(a->emin.year > b->emin.year)
				return 1;
			if(a->emin.year < b->emin.year)
				return -1;

			if(a->emin.month > b->emin.month)
				return 1;
			if(a->emin.month < b->emin.month)
				return -1;

			if(a->emin.day > b->emin.day)
				return 1;
			if(a->emin.day < b->emin.day)
				return -1;
		}
	}

	if(a->smin.day > b->smin.day)
		return 1;
	if(a->smin.day < b->smin.day)
		return -1;

	if(a->smin.minute == 0xffff && b->smin.minute != 0xffff)
		return -1;
	if(a->smin.minute != 0xffff && b->smin.minute == 0xffff)
		return 1;
	if(a->smin.minute < b->smin.minute)
		return -1;
	if(a->smin.minute > b->smin.minute)
		return 1;

	if(a->melody[0] && !b->melody[0])
		return -1;
	if(!a->melody[0] && b->melody[0])
		return 1;

	return strcmp(a->data[SCL_TITLE].buf, b->data[SCL_TITLE].buf);
}

DEFINE_ADD(scl, scl_holiday, h)

DEFINE_DEL(scl, scl_holiday, h)

DEFINE_ADD(scl, scl_holiday, w)

DEFINE_DEL(scl, scl_holiday, w)

DEFINE_QSORT(scl_holiday)

void
scl_holidayh_sort(scl_t *db)
{
	scl_holiday_qsort(db->hhead, db->htail);

	return;
}

void
scl_holidayw_sort(scl_t *db)
{
	scl_holiday_qsort(db->whead, db->wtail);

	return;
}

int
scl_holiday_cmp(scl_holiday_t *a, scl_holiday_t *b)
{
	if(a->year > b->year)
		return 1;
	if(a->year < b->year)
		return -1;

	if(a->month > b->month)
		return 1;
	if(a->month < b->month)
		return -1;

	return 0;
}
