/* $Id: xps.c,v 1.14 2006/08/07 15:55:51 geni Exp $
 *
 * Copyright (c) 2005 Huidae Cho
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "onhand.h"
#include "define.h"

static void xps_rec_qsort(xps_rec_t *head, xps_rec_t *tail);

int
xps_read(xps_t *db, char *file)
{
	int size, i;
	u8 *data, *ptr, *ptr2;
	u32 offset, offset2, offset3;
	xps_rec_t *rec;

	if(read_file(file, &data, &size)){
		ERROR("read_file");
		return -1;
	}
	/* the minimum size is 32 = 4+4+2+2+2+2+12+4 */
	if(size < 32 || memcmp(data, pim_magic, 4)){
		ERROR("%s: not an xps file", file);
		return -1;
	}

	xps_init(db);

	memcpy(&db->group, data+8, 2);
	if(db->group != 0xffff)
		db->group -= 0x65;
	memcpy(&db->currency, data+10, 2);
	db->currency -= 0xc9;
	memcpy(&db->display, data+14, 2);

	memcpy(&offset, data+28, 4);
	while(offset != 0xffffffff && offset < size){
		ptr = data + offset;
		memcpy(&offset, ptr+2, 4);
		memcpy(&offset2, ptr+6, 4);
		while(offset2 != 0xffffffff && offset2 < size){
			ptr = data + offset2;
			memcpy(&offset2, ptr+2, 4);
			for(i = 0; i < 31; i++){
				memcpy(&offset3, ptr+874+4*i, 4);
				while(offset3 != 0xffffffff && offset3 < size){
					ptr2 = data + offset3;
					memcpy(&offset3, ptr2+4, 4);
					if(ptr2[8])
						continue;
					rec = (xps_rec_t *)zalloc(
							sizeof(xps_rec_t));
					rec->date.day = ptr2[16];
					rec->date.month = ptr2[17];
					memcpy(&rec->date.year, ptr2+18, 2);
					memcpy(&rec->currency, ptr2+20, 2);
					rec->currency -= 0xc9;
					memcpy(&rec->expense, ptr2+22, 4);
					memcpy(&rec->group, ptr2+28, 2);
					rec->group -= 0x65;
					memcpy(&rec->memo, ptr2+30, 34);
					rec->memo[34] = 0;
					rec->prev = rec->next = NULL;
					if(db->head){
						rec->prev = db->tail;
						db->tail->next = rec;
						db->tail = rec;
					}else
						db->head = db->tail = rec;
				}
			}
		}
	}

	free(data);

	xps_rec_sort(db);

	return 0;
}

int
xps_write(xps_t *db, char *file)
{
	int fd, i, j;
	u8 i8;
	u16 i16;
	u32 i32, offset, offset2;
	date_t date;
	xps_rec_t *rec, *r;

	_open();

	_write(pim_magic, 4);
	i32 = 0;
	_write(&i32, 4);
	if(db->group == 0xffff)
		i16 = 0xffff;
	else
		i16 = db->group + 0x65;
	_write(&i16, 2);
	i16 = db->currency + 0xc9;
	_write(&i16, 2);
	i16 = 1;
	_write(&i16, 2);
	_write(&db->display, 2);
	i32 = 2;
	_write(&i32, 4);
	i32 = 0;
	for(i = 0; i < 2; i++)
		_write(&i32, 4);

	/* header */
	if(db->head)
		offset = 32;
	else
		offset = 0xffffffff;

	_write(&offset, 4);

	date.year = 0;
	for(rec = db->head; rec; rec = rec->next){
		if(date.year != rec->date.year){
			date.year = rec->date.year;
			/* year header */
			/* year */
			_write(&date.year, 2);

			offset += 10;
			offset2 = offset;
			date.month = 0;
			for(r = rec; r && date.year == r->date.year;
					r = r->next){
				if(date.month != r->date.month){
					date.month = r->date.month;
					offset2 += 998;
				}
				offset2 += 64;
			}
			if(!r)
				offset2 = 0xffffffff;
			/* offset for next year */
			_write(&offset2, 4);

			/* offset for the first month of the year */
			_write(&offset, 4);

			date.month = 0;
		}
		if(date.month != rec->date.month){
			/* month header */
			/* month */
			date.month = rec->date.month;
			_write(&date.month, 1);
			i8 = 0;
			_write(&i8, 1);

			offset += 998;
			offset2 = offset;
			if(date.month < 12){
				for(r = rec; r && date.year == r->date.year &&
						date.month == r->date.month;
						r = r->next)
					offset2 += 64;
				if(!r)
					offset2 = 0xffffffff;
			}else
				offset2 = 0xffffffff;
			/* offset for next month */
			_write(&offset2, 4);

			/* daily total */
			r = rec;
			for(j = 0; j < 7; j++){
			for(i = 1; i < 32; i++){
				i32 = 0;
				for(; r && date.year == r->date.year &&
					date.month == r->date.month &&
					i == r->date.day && j == r->group;
					r = r->next)
					i32 += r->expense;
				_write(&i32, 4);
			}
			}
			/* offset for daily records */
			offset2 = offset;
			r = rec;
			for(i = 1; i < 32; i++){
				for(; r && date.year == r->date.year &&
					date.month == r->date.month &&
					i > r->date.day; r = r->next)
					offset2 += 64;
				if(r && date.year == r->date.year &&
					date.month == r->date.month &&
					i == r->date.day)
					i32 = offset2;
				else
					i32 = 0xffffffff;
				_write(&i32, 4);
			}

			date.day = 0;
		}

		/* prev */
		if(date.day != rec->date.day){
			date.day = rec->date.day;
			i32 = 0xffffffff;
		}else
			i32 = offset - 64;
		_write(&i32, 4);

		offset += 64;

		/* next */
		if(rec->next && date.year == rec->next->date.year &&
				date.month == rec->next->date.month &&
				date.day == rec->next->date.day)
			i32 = offset;
		else
			i32 = 0xffffffff;
		_write(&i32, 4);

		/* overwritten = 0, unknown [7] */
		i32 = 0;
		_write(&i32, 4);
		/* TODO: 00 ... 00 for unknown works! */
		i32 = 0;
		_write(&i32, 4);

		/* data */
		_write(&rec->date.day, 1);
		_write(&rec->date.month, 1);
		_write(&rec->date.year, 2);
		i16 = rec->currency + 0xc9;
		_write(&i16, 2);
		_write(&rec->expense, 4);
		i16 = 0;
		_write(&i16, 2);
		i16 = rec->group + 0x65;
		_write(&i16, 2);
		_write(rec->memo, 34);
	}

	_close();

	return 0;
}

void
xps_scan(xps_t *db, FILE *fp)
{
	char *line;
	u16 i16;
	date_t d;
	xps_rec_t *rec;

	rec = NULL;
	while((line = read_line(fp))){
		if(!line[0]){
			if(rec){
				xps_rec_add(db, rec);
				rec = NULL;
			}
		}else
		if(strstr(line, "default group: ") == line){
			if(sscanf(line+15, "%hd", &i16) == 1){
			if(i16 > 0 && i16 < 7)
				db->group = i16;
			}
		}else
		if(strstr(line, "default currency: ") == line){
			if(sscanf(line+18, "%hd", &i16) == 1){
			if(i16 > 0 && i16 < 7)
				db->currency = i16;
			}
		}else
		if(strstr(line, "beginning display: ") == line){
			if(!strcmp(line+19, "monthly list"))
				db->display = 0;
			else
			if(!strcmp(line+19, "daily list"))
				db->display = 1;
			else
			if(!strcmp(line+19, "input"))
				db->display = 2;
		}else
		if(strstr(line, "date: ") == line){
			if(sscanf(line+6, "%04hd-%02hhd-%02hhd", &d.year,
						&d.month, &d.day) == 3){
			if(!rec)
				rec = (xps_rec_t *)zalloc(sizeof(xps_rec_t));
			rec->date.year = d.year;
			rec->date.month = d.month;
			rec->date.day = d.day;
			}
		}else
		if(strstr(line, "group: ") == line){
			if(sscanf(line+7, "%hd", &i16) == 1){
			if(!rec)
				rec = (xps_rec_t *)zalloc(sizeof(xps_rec_t));
			rec->group = i16;
			}
		}else
		if(strstr(line, "currency: ") == line){
			if(sscanf(line+10, "%hd", &i16) == 1){
			if(!rec)
				rec = (xps_rec_t *)zalloc(sizeof(xps_rec_t));
			rec->currency = i16;
			}
		}else
		if(strstr(line, "expense: ") == line){
			int a, b;

			if(sscanf(line+9, "%d.%02d", &a, &b) == 2){
			if(!rec)
				rec = (xps_rec_t *)zalloc(sizeof(xps_rec_t));
			rec->expense = a * 100 + b;
			}
		}else
		if(strstr(line, "memo: ") == line){
			if(!rec)
				rec = (xps_rec_t *)zalloc(sizeof(xps_rec_t));
			strncpy(rec->memo, line+6, 34);
			rec->memo[34] = 0;
			sub_chr(rec->memo, '\t', '\n');
		}
		free(line);
	}

	return;
}

void
xps_print(xps_t *db, FILE *fp)
{
	xps_rec_t *rec;

	fprintf(fp, "default group: %d\n", db->group);
	fprintf(fp, "default currency: %d\n", db->currency);
	fprintf(fp, "beginning display: %s\n",
			(db->display == 0 ? "monthly list" :
			 (db->display == 1 ? "daily list" : "input")));
	fprintf(fp, "\n");
	for(rec = db->head; rec; rec = rec->next){
		fprintf(fp, "date: %04d-%02d-%02d %s\n", rec->date.year,
				rec->date.month, rec->date.day,
				dow[day_of_week(rec->date.year, rec->date.month,
					rec->date.day)]);
		fprintf(fp, "group: %d\n", rec->group);
		fprintf(fp, "currency: %d\n", rec->currency);
		fprintf(fp, "expense: %d.%02d\n", rec->expense/100,
				rec->expense%100);
		fprintf(fp, "memo: ");
		print_str(fp, rec->memo);
		fprintf(fp, "\n");
		fprintf(fp, "\n");
	}

	return;
}

void
xps_init(xps_t *db)
{
	db->group = db->currency = db->display = 0;
	db->head = db->tail = NULL;

	return;
}

void
xps_free(xps_t *db)
{
	xps_rec_t *rec, *rec_next;

	for(rec = db->head; rec; rec = rec_next){
		rec_next = rec->next;
		xps_rec_free(rec);
	}

	return;
}

void
xps_rec_free(xps_rec_t *rec)
{
	free(rec);

	return;
}

DEFINE_ADD(xps, xps_rec, )

DEFINE_DEL(xps, xps_rec, )

DEFINE_QSORT(xps_rec)

void
xps_rec_sort(xps_t *db)
{
	xps_rec_qsort(db->head, db->tail);

	return;
}

int
xps_rec_cmp(xps_rec_t *a, xps_rec_t *b)
{
	if(a->date.year > b->date.year)
		return 1;
	if(a->date.year < b->date.year)
		return -1;

	if(a->date.month > b->date.month)
		return 1;
	if(a->date.month < b->date.month)
		return -1;

	if(a->date.day > b->date.day)
		return 1;
	if(a->date.day < b->date.day)
		return -1;

	if(a->group > b->group)
		return 1;
	if(a->group < b->group)
		return -1;

	if(a->currency > b->currency)
		return 1;
	if(a->currency < b->currency)
		return -1;

	return 0;
#if 0
	if(a->expense < b->expense)
		return 1;
	if(a->expense > b->expense)
		return -1;

	return strcmp(a->memo, b->memo);
#endif
}
